import numpy as np
from gibson_dataset.utilities.data_pipeline import DataPipeline
from typing import Dict, Union, TypeVar, Callable, Any
from threading import Lock
from functools import update_wrapper, wraps

AnotherActivePipelineException = Exception

TCallable = TypeVar('TCallable', bound=Callable[..., Any])


def synchronized_on_field(field_name: str, raise_expection: bool = True) -> Callable[[TCallable], TCallable]:
    def decorator(method: TCallable) -> TCallable:
        @wraps(method)
        def sync_method(self, *args, **kwargs):
            lock = self._locks[field_name]
            with lock:
                if raise_expection and self._pipelines[field_name] is not None:
                    raise AnotherActivePipelineException('Be careful, there is another active pipeline for {0}, please terminate it.'.format(field_name))
                return method(self, *args, **kwargs)
        return sync_method
    return decorator


class Sample:
    """
    This class represents a sample of the dataset. A sample is composed by the colored image, the relative depth data and the generated depth image.
    For manipulating each field, a pipeline can be created. There can only be one active pipeline for each field simultaneously.
    A pipeline is considered terminated when get_data() method id called.
    """
    def __init__(self, color_image: np.ndarray, depth_data: np.ndarray):
        self._color_image: np.ndarray = color_image
        self._depth_data: np.ndarray = depth_data
        self._depth_image: np.ndarray = np.empty((0, 0, 0), dtype=np.uint8)

        self._locks: Dict[str, Lock] = {'color_image': Lock(), 'depth_image': Lock(), 'depth_data': Lock()}
        self._pipelines: Dict[str, Union[DataPipeline, None]] = {'color_image': None, 'depth_image': None, 'depth_data': None}

    @synchronized_on_field(field_name='color_image', raise_expection=True)
    def create_pipeline_for_color_image(self, use_gpu: bool = False) -> DataPipeline:
        """
        Create and return a new pipeline to elaborate color_image. If there is another active pipeline, raise an AnotherActivePipelineException
        :raise AnotherActivePipelineException
        :param use_gpu: if this param is true, the pipeline is executed in GPU
        :return: a new pipeline instance
        :rtype DataPipeline
        """
        def assign(data: np.array) -> np.array:
            self._color_image = data
            self._pipelines['color_image'] = None
            return data

        self._pipelines['color_image'] = DataPipeline(data=self._color_image, use_gpu=use_gpu, end_function=assign)
        return self._pipelines['color_image']

    @synchronized_on_field(field_name='depth_data', raise_expection=True)
    def create_pipeline_for_depth_data(self, use_gpu: bool = False) -> DataPipeline:
        """
        Create and return a new pipeline to elaborate depth data. If there is another active pipeline, raise an AnotherActivePipelineException.
        :raise AnotherActivePipelineException
        :param use_gpu: if this param is true, the pipeline is executed in GPU
        :type use_gpu: bool
        :return: a new pipeline instance
        :rtype DataPipeline
        """
        def assign(data: np.array) -> np.array:
            self._depth_data = data
            self._pipelines['depth_data'] = None
            return data

        self._pipelines['depth_data'] = DataPipeline(data=self._depth_data, use_gpu=use_gpu, end_function=assign)
        return self._pipelines['depth_data']

    @synchronized_on_field(field_name='depth_image', raise_expection=True)
    def create_pipeline_for_depth_image(self, use_gpu: bool = False) -> DataPipeline:
        """
        Create and return a new pipeline to elaborate depth image. If there is another active pipeline, raise an AnotherActivePipelineException.
        :raise AnotherActivePipelineException
        :param use_gpu: if this param is true, the pipeline is executed in GPU
        :type use_gpu: bool
        :return: a new pipeline instance
        :rtype DataPipeline
        """
        def assign(data: np.array) -> np.array:
            self._depth_data = data
            self._pipelines['depth_image'] = None
            return data

        self._pipelines['depth_image'] = DataPipeline(data=self._depth_image, use_gpu=use_gpu, end_function=assign)
        return self._pipelines['depth_image']

    @synchronized_on_field(field_name='depth_image', raise_expection=True)
    def create_pipeline_to_generate_depth_image(self, limit: float = 10, use_gpu: bool = False):
        """
        Creates a pipeline to generate depth image starting from the depth data.
        The pipeline has all the necessary operations to create the depth image, you just need to run it and get the data.
        If there is another active pipeline, raise an AnotherActivePipelineException.
        :param limit: the new limit used to scale the array's values
        :type limit: float
        :param use_gpu: if this param is true, the pipeline is executed in GPU
        :type use_gpu: bool
        :return: the pipeline
        :rtype: DataPipeline
        """
        def assign(data: np.array) -> np.array:
            self._depth_image = data
            self._pipelines['depth_image'] = None
            return data

        def limit_range(data):
            data[data > limit] = limit
            return data

        self._pipelines['depth_image'] = DataPipeline(data=self._depth_data, use_gpu=use_gpu, end_function=assign)\
            .add_operation(function=limit_range).scale_values_on_new_max(new_max=255) \
            .add_operation(function=lambda data: data.astype('uint8'))

        return self._pipelines['depth_image']

    @synchronized_on_field(field_name='color_image', raise_expection=True)
    def get_color_image(self) -> np.ndarray:
        """
        Return color_image. If there is an active pipeline for this field, raises an AnotherActivePipelineException.
        :return: the colored image
        :rtype: np.ndarray
        """
        return self._color_image

    @synchronized_on_field(field_name='depth_data', raise_expection=True)
    def get_depth_data(self) -> np.ndarray:
        """
        Return depth data. If there is an active pipeline for this field, raises an AnotherActivePipelineException.
        :return: the depth data
        :rtype: np.ndarray
        """
        return self._depth_data

    @synchronized_on_field(field_name='depth_image', raise_expection=True)
    def get_depth_image(self) -> np.ndarray:
        """
        Return depth_image. If there is an active pipeline for this field, raises an AnotherActivePipelineException.
        :return: the depth image
        :rtype: np.ndarray
        """
        return self._depth_image

    @synchronized_on_field(field_name='color_image', raise_expection=False)
    def get_pipeline_color_image(self) -> DataPipeline:
        return self._pipelines['color_image']

    @synchronized_on_field(field_name='depth_image', raise_expection=False)
    def get_pipeline_depth_image(self) -> DataPipeline:
        return self._pipelines['depth_image']

    @synchronized_on_field(field_name='depth_data', raise_expection=False)
    def get_pipeline_depth_data(self) -> DataPipeline:
        return self._pipelines['depth_data']



from typing import List


class Color:
    """
    This class define a color using red, green, blue value
    """
    def __init__(self, red: int, green: int, blue: int):
        self._red = red
        self._green = green
        self._blue = blue

    def RGB(self) -> List[int]:
        """
        This methods create a RGB representation of the color

        :return: a list of int, containing the color values in RGB format
        """
        return [self._red, self._green, self._blue]

    def BGR(self) -> List[int]:
        """
        This methods create a BGR representation of the color

        :return: a list of int, containing the color values in BGR format
        :rtype: list[int]
        """
        return [self._blue, self._green, self._red]


import queue
from typing import Callable, Any

import numpy as np
import gibson_dataset.utilities.engine_selector as eg


class DataPipeline:
    """
    This class constructs an elaboration pipeline to elaborate a single data.
    A pipeline is composed by a series of consecutive operations performed to the same data.
    A pipeline can be executed in CPU or GPU (using CuPy)
    :
    """

    def __init__(self, data: np.ndarray, use_gpu: bool, end_function: Callable[[np.ndarray], np.ndarray] = lambda data: data):
        """
        Creates a new pipeline
        :param data: the data to be modified
        :param use_gpu: indicates of the pipeline must be executed in gpu
        :param end_function: a function which can be used to automatically save the data outside the pipeline
        """
        self._data = data
        self._use_gpu = use_gpu
        self._end_function = end_function
        self._operations = queue.Queue()

        self._xp = eg.get_engine(eg.NUMPY if not use_gpu else eg.CUPY)

        # IF the pipeline is executed in gpu, transfer data to device mandatory
        if use_gpu:
            self._cuda_stream = self._xp.cuda.Stream(non_blocking=True)
            self._transfer_data_to_device()

    def _transfer_data_to_device(self) -> 'DataPipeline':
        """
        Adds to the pipeline's operations the data migration to GPU
        :return: return the pipeline
        :rtype DataPipeline
        """
        self._operations.put(lambda data: self._xp.asarray(data))

        return self

    def _get_data_from_device(self) -> 'DataPipeline':
        """
        Obtains the data from GPU
        :return: return the pipeline
        :rtype DataPipeline
        """
        self._operations.put(lambda data: self._xp.asnumpy(data, stream=self._cuda_stream))

        return self

    def run(self) -> 'DataPipeline':
        """
        Runs the pipeline returning the modified data.
        Note that if the pipeline uses GPU, this is an ASYNC operation.
        :return: the modified data
        :rtype np.ndarray
        """
        if self._use_gpu:
            self._get_data_from_device()

        ret = self._data

        if self._use_gpu:
            self._cuda_stream.use()

        for operation in self._operations.queue:
            ret = operation(ret)

        self._data = ret

        return self

    def get_data(self) -> np.ndarray:
        """
        Returns the data after the pipeline execution finishes and call the end-function specified in the constructor.
        This a SYNC operation, so if the pipeline is running using GPU,
        the current thread is blocked until all operations in the pipeline are completed.
        :return: np.ndarray
        :rtype np.ndarray
        """
        if self._use_gpu:
            self._cuda_stream.synchronize()

        return self._end_function(self._data)

    def convert_bgr_to_rgb(self) -> 'DataPipeline':
        """
        Adds to the pipeline's operations the conversion from BGR to RGB
        :return: the pipeline
        :rtype DataPipeline
        """
        self._operations.put(lambda data: data[:, :, [2, 1, 0]])

        return self

    def convert_rgb_to_bgr(self) -> 'DataPipeline':
        """
        Adds to the pipeline's operations the conversion from RGB to BGR
        :return: the pipeline
        :rtype DataPipeline
        """
        self._operations.put(lambda data: data[:, :, ::-1])

        return self

    def around(self, decimals: int = 0) -> 'DataPipeline':
        """
        Round an array of float to the given number of decimals
        :param decimals: the number of decimal places to round
        :type decimals: int
        :return: the pipeline
        :rtype DataPipeline
        """
        self._operations.put(lambda data: self._xp.around(data, decimals=decimals))

        return self

    def scale_values_on_new_max(self, new_max: float) -> 'DataPipeline':
        """
        Scales all elements of an array to a new maximum value. The min value is zero.
        Tipically used to produce a visualizable depth image from depth data
        :param new_max:
        :type new_max: float
        :return:
        """
        self._operations.put(lambda data: data * (new_max / self._xp.nanmax(data)))

        return self


    def add_operation(self, function: Callable[[Any], Any]) -> 'DataPipeline':
        """
        Adds to the pipeline operations queue a custom operation defined by the used. Typically used to add indexing operation: data[data > 10] = 1
        :param function: a function that accpet a NumPy or a CuPy array and returns one of them
        :type function: Callable[[Any], Any]
        :return: the pipeline
        :rtype DataPipeline
        """
        self._operations.put(function)

        return self



import numpy

Engine = str
NUMPY: Engine = 'numpy'
CUPY: Engine = 'cupy'


def check_cuda_support():
    """
    Print a message which tell if the current machine supports CuPy
    :return: None
    """
    try:
        import cupy as cp
        print('Success: you have CUDA Toolkit correctly installed, so you can use CuPy :)')
    except Exception as e:
        print('Something went wrong, you cant use CuPy :( \n' + str(e))


def get_engine(engine: Engine):
    """

    :raise Exception if cupy is not supported
    :param engine: the engine to return
    :type engine: Engine
    :return:
    """
    if engine == NUMPY:
        return numpy
    else:
        try:
            import cupy
            return cupy
        except Exception as e:
            raise Exception('Something went wrong, you cant use GPU :( \n' + str(e))
